;;; util.scm: Utility functions for uim.
;;;
;;; Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
;;; You can use, modify, distribute under the term of GPL or BSD Licence.

;;
(define alphabet-char?
  (lambda (c)
    (and (number? c)
	 (or
	  (and (> c 64) (< c 91))
	  (and (> c 96) (< c 123))))))
;;
(define usual-char?
  (lambda (c)
    (and (number? c)
	 (or
	  (and (> c 32) (< c 127))))))
;;
(define numeral?
  (lambda (c)
    (and (number? c)
	  (and (> c 47) (< c 59)))))

;;
(define to-lower-char
  (lambda (c)
    (if (and (alphabet-char? c) (< c 91))
	(+ c 32)
	c)))
;;
(define string-list-concat
  (lambda (lst)
    (let ((len (length lst)))
      (if (= len 0)
	  ""
	  (string-append
	   (string-list-concat (cdr lst))
	   (car lst))))))
;;
(define string-find
  (lambda (lst str)
    (if lst
	(if
	 (string-equal? (car lst) str)
	 #t
	 (string-find (cdr lst) str))
	())))
;;
(define truncate-list
  (lambda (lst len)
    (if (< (length lst) len)
	nil
	(if (= len 0)
	    '()
	    (cons (car lst)
		  (truncate-list
		   (cdr lst)
		   (- len 1)))))))

;; procedural 'or' for use with 'apply'
;; e.g. (apply proc-or boolean-lst)
;; should be deprecated and replaced with a proper, Schemer's way
(define proc-or
  (lambda xs
    (if (null? xs)
	#f
	(or (car xs)
	    (apply proc-or (cdr xs))))))

;;
;; R5RS procedures (don't expect 100% compatibility)
;;
(define else #t)

(define boolean?
  (lambda (x)
    (or (eq? x #t)
        (eq? x #f))))

(define list?
  (lambda (x)
    (or (null? x)
	(and (pair? x)
	     (list? (cdr x))))))

;; update style-element vars
;; style-spec requires list of (style-element-name . validator)
(define update-style
  (lambda (style-spec style)
    (let* ((elem (car style))
	   (name (car elem))
	   (val (if (symbol? (cdr elem))
		    (symbol-value (cdr elem))
		    (cdr elem)))
	   (spec (assq name style-spec))
	   (valid? (symbol-value (cdr spec))))
      (if (valid? val)
	  (set-symbol-value! name val))
      (if (not (null? (cdr style)))
	  (update-style style-spec (cdr style))))))

;;
;; Preedit color related configurations and functions.
;;
(define reversed-preedit-foreground)
(define reversed-preedit-background)
(define separator-foreground)
(define separator-background)
(define reversed-separator-foreground)
(define reversed-separator-background)

(define uim-color 'uim-color-uim)
(define uim-color-spec
  '((reversed-preedit-foreground   . string?)
    (reversed-preedit-background   . string?)
    (separator-foreground          . string?)
    (separator-background          . string?)
    (reversed-separator-foreground . string?)
    (reversed-separator-background . string?)))

;; predefined color styles
(define uim-color-uim
  '((reversed-preedit-foreground   . "white")
    (reversed-preedit-background   . "black")
    (separator-foreground          . "lightsteelblue")
    (separator-background          . "")
    (reversed-separator-foreground . "white")
    (reversed-separator-background . "black")))
(define uim-color-atok
  '((reversed-preedit-foreground   . "black")
    (reversed-preedit-background   . "blue")
    (separator-foreground          . "lightsteelblue")
    (separator-background          . "")
    (reversed-separator-foreground . "black")
    (reversed-separator-background . "blue")))
